/*
   VRExperience.js
 */
import {Vector3D, Matrix4D} from "./libraries/LinearAlgebra.js";
import {Entity} from "./entities/Entity.js";
import {CompositeEntity} from "./entities/CompositeEntity.js";
import {GamepadEntity} from "./entities/GamepadEntity.js";
import {Scene} from "./scenes/Scene.js";
export function startExperience(canvas)
{
    var gl = null;
    var entity = null;
    var scene = null;
    var vr_display = null;
    var buffers = null;
    var vr_present_button;
    function init(preserve_drawing_buffer)
    {
        var gl_attribs = {
            alpha: true,
            antialias: true,
            preserveDrawingBuffer: preserve_drawing_buffer
        };
        gl = canvas.getContext("webgl", gl_attribs);
        canvas.getContext('webgl', gl_attribs) ||
            canvas.getContext('experimental-webgl', gl_attribs);
        // If we don't have a GL context, give up now
        if (!gl) {
            alert('Unable to initialize WebGL. Your browser or machine '+
                'may not support it.');
            return;
        }
        scene = new Scene(gl);
        var spin_box;
        spin_box = new Entity(null, [1,0,0], [.1,.2,.3], .5);
        spin_box.axis = Vector3D.create([0, 0, 1]);
        spin_box.frequency = 1;
        scene.addEntity(spin_box);
        var box = new Entity(null, [-2,0,0], [.3,.2,.1], .5);
        scene.addEntity(box);
        //scene.view_entity = spin_box;
        //spin_box.allow_rendering = false;
        //spin_box.orientation_tracking = true;
        // Wait until we have a WebGL context to resize and start rendering.
        window.addEventListener("resize", onResize, false);
        onResize();
        window.requestAnimationFrame(animate);
    }
    function onResize() {
        if (vr_display && vr_display.isPresenting) {
            var left_eye = vr_display.getEyeParameters("left");
            var right_eye = vr_display.getEyeParameters("right");
            canvas.width = Math.max(left_eye.renderWidth,
                right_eye.renderWidth) * 2;
            canvas.height = Math.max(left_eye.renderHeight,
                right_eye.renderHeight);
        } else {
            canvas.width =
                canvas.offsetWidth;
            canvas.height =
                canvas.offsetHeight;
        }
    }
    var then = 0;
    // Draw the scene repeatedly
    function onContextLost(event)
    {
        event.preventDefault();
        console.log( 'WebGL Context Lost.' );
        gl = null;
    }
    function onContextRestored(event)
    {
        console.log( 'WebGL Context Restored.' );
        init(vr_display ? vr_display.capabilities.hasExternalDisplay : false);
    }
    function onVRRequestPresent()
    {
        vr_display.requestPresent([{ source: canvas }]).then(function () {
            console.log("hi there!! presenting");
        }, function (err) {
            var errMsg = "requestPresent failed.";
            if (err && err.message) {
                errMsg += "<br />" + err.message
            }
            console.log(errMsg);
        });
    }
    function onVRExitPresent()
    {
        if (!vr_display.isPresenting) {
            return;
        }
        vr_display.exitPresent().then(function () {
            window.requestAnimationFrame(animate);
        }, function () {
            console.log("exitPresent failed.");
        });
    }
    function onVRPresentChange()
    {
        onResize();
        if (vr_display.isPresenting) {
            if (vr_display.capabilities.hasExternalDisplay) {
                removeButton(vr_present_button);
                vr_present_button = addButton("Exit VR",
                    onVRExitPresent);
            }
        } else {
            if (vr_display.capabilities.hasExternalDisplay) {
                removeButton(vr_present_button);
                vr_present_button = addButton("Enter VR",
                    onVRRequestPresent);
            }
        }
    }
    canvas.addEventListener('webglcontextlost', onContextLost, false );
    canvas.addEventListener('webglcontextrestored', onContextRestored,
        false);
    if (navigator.getVRDisplays) {
        navigator.getVRDisplays().then(function (displays) {
            if (displays.length > 0) {
                vr_display = displays[displays.length - 1];
                init(true);
                if (vr_display.capabilities.canPresent) {
                    vr_present_button = addButton("Enter VR",
                        onVRRequestPresent);
                }
                window.addEventListener('vrdisplaypresentchange',
                    onVRPresentChange, false);
                window.addEventListener('vrdisplayactivate', onVRRequestPresent,
                    false);
                window.addEventListener('vrdisplaydeactivate', onVRExitPresent,
                    false);
            } else {
                init(false);
                window.requestAnimationFrame(animate);
                console.log("No VRDisplays found.");
            }
        },
        function () {
            init(false);
            window.requestAnimationFrame(animate);
            console.log("No WebVR support.");
        });
    }
    /*
     */
    function animate(now)
    {
        now *= 0.001;  // convert to seconds
        var delta_time = now - then;
        then = now;
        scene.update(delta_time);
        renderExperience(vr_display, scene);
        if (vr_display && vr_display.isPresenting) {
            vr_display.requestAnimationFrame(animate);
        } else {
            window.requestAnimationFrame(animate);
        }
    }
}
/*
  Draw the VR Experience.
 */
function renderExperience(vr_display, scene)
{
    var gl = scene.gl;
    gl.clearColor(0.0, 0.0, 0.0, 1.0);  // Clear to black, fully opaque
    gl.clearDepth(1.0);                 // Clear everything
    gl.enable(gl.DEPTH_TEST);           // Enable depth testing
    gl.depthFunc(gl.LEQUAL);            // Near things obscure far things
    // Clear the canvas before we start drawing on it.
    gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
    if (vr_display && vr_display.isPresenting) {
        var frame_data = new VRFrameData();
        vr_display.getFrameData(frame_data);
        scene.orientation = frame_data.pose;
        scene.projection = frame_data.leftProjectionMatrix;
        gl.viewport(0, 0, gl.canvas.clientWidth * 0.5, gl.canvas.clientHeight);
        scene.render();
        scene.projection = frame_data.rightProjectionMatrix;
        gl.viewport(gl.canvas.clientWidth * 0.5, 0,
            gl.canvas.clientWidth * 0.5, gl.canvas.clientHeight);
        scene.render();
        vr_display.submitFrame();
    } else {
        scene.render();
    }
}
function addButtonElement(message)
{
    var button_elt = document.createElement("div");
    var webgl_canvas = document.getElementById('glcanvas');
    var button_container = document.getElementById("vr-button-container");
    if (!button_container) {
        button_container = document.createElement("div");
        button_container.id = "vr-button-container";
        Object.assign(button_container.style, {
            fontFamily : "sans-serif",
            position : "absolute",
            zIndex : "999",
            left : "0",
            bottom : "0",
            right : "0",
            margin : "0",
            padding : "0"
        });
        button_container.align = "right";
        webgl_canvas.parentNode.appendChild(button_container);
    }
    Object.assign(button_elt.style, {
        color: "#FFF",
        fontWeight: "bold",
        backgroundColor : "#888",
        borderRadius : "5px",
        border : "3px solid #555",
        position : "relative",
        display :  "inline-block",
        margin : "0.5em",
        padding : "0.75em",
        cursor : "pointer",
    });
    button_elt.align = "center";
    button_elt.innerHTML = "<img src='vricon.png' style='width:1in' /><br />" +
        message;
    button_container.appendChild(button_elt);
    return button_elt;
}
function addButton(message, callback)
{
    var element = addButtonElement(message);
    element.addEventListener("click", function (event) {
        callback(event);
        event.preventDefault();
    }, false);
    return {
        element: element,
    };
}
function removeButton(button)
{
    if (!button) {
        return;
    }
    if (button.element.parentElement) {
        button.element.parentElement.removeChild(button.element);
    }
}
